@@ -19,32 +19,32 @@
  *
  */
 
+#ifdef WIN32
+#include <windows.h>
+#endif /* WIN32 */
+
 #include "logo.h"
 /*   #include "globals.h"   has been moved further down */
 #include <math.h>
 
 #ifdef mac
 #include "macterm.h"
-#else
-#ifdef ibm
-#ifdef __ZTC__
+#elif defined(WIN32)
+#include "win32trm.h"
+#elif defined(__ZTC__)
 #include <fg.h>
 #include "ztcterm.h"
-#else
-#include "ibmterm.h"
-#endif
-#else
-#ifdef x_window
+#elif defined(x_window)
 #include "xgraphics.h"
+#elif defined(ibm)
+#include "ibmterm.h"
 #else
 #include "nographics.h"
-#endif
-#endif
-#endif
+#endif /* end this whole big huge tree */
 
 #include "globals.h"
 
-#ifdef __ZTC__
+#if defined(__ZTC__) && !defined(WIN32) /* sowings */
 #define total_turtle_bottom_max (-(MaxY/2))
 #else
 #define total_turtle_bottom_max turtle_bottom_max
@@ -64,12 +64,23 @@
 /* NOTE: See the files (macterm.c and macterm.h) or (ibmterm.c and ibmterm.h)
    for examples of the functions and macros that this file assumes exist. */
 
+#define One (sizeof(int))
+#define Two (2*One)
+#define Three (3*One)
+
+#define PENMODE_PAINT	0
+#define PENMODE_ERASE	1
+#define PENMODE_REVERSE	2
+
+int internal_penmode = PENMODE_PAINT;
+
 mode_type current_mode = wrapmode;
 FLONUM turtle_x = 0.0, turtle_y = 0.0, turtle_heading = 0.0;
-BOOLEAN turtle_shown = TRUE;
 FLONUM x_scale = 1.0, y_scale = 1.0;
+BOOLEAN turtle_shown = TRUE;
 FLONUM wanna_x = 0.0, wanna_y = 0.0;
 BOOLEAN out_of_bounds = FALSE;
+void setpos_bynumber(FLONUM, FLONUM);
 
 char record[GR_SIZE];
 int record_index = 0;
@@ -86,8 +97,7 @@ double pfmod(double x, double y) {
     return temp;
 }
 
-FLONUM cut_error(FLONUM n)
-{
+FLONUM cut_error(FLONUM n) {
     n *= 1000000;
     n = (n > 0 ? floor(n) : ceil(n));
     n /= 1000000;
@@ -95,8 +105,7 @@ FLONUM cut_error(FLONUM n)
     return(n);
 }
 
-FIXNUM g_round(FLONUM n)
-{
+FIXNUM g_round(FLONUM n) {
     n += (n < 0.0 ? -0.5 : 0.5);
     if (n < 0.0)
 	return((FIXNUM)ceil(n));
@@ -104,9 +113,11 @@ FIXNUM g_round(FLONUM n)
 }
 
 /************************************************************/
+void draw_turtle_helper(void);
+void check_x_high(void);
+void check_x_low(void);
 
-void draw_turtle()
-{
+void draw_turtle(void) {
     if (!turtle_shown) return;
     draw_turtle_helper();
     /* all that follows is for "turtle wrap" effect */
@@ -130,8 +141,7 @@ void draw_turtle()
     check_x_low();
 }
 
-check_x_high()
-{
+void check_x_high(void) {
     if ((turtle_x > turtle_right_max - turtle_height) &&
 	    (current_mode == wrapmode)) {
 	turtle_x -= (screen_width + 1);
@@ -140,8 +150,7 @@ check_x_high()
     }
 }
 
-check_x_low()
-{
+void check_x_low(void) {
     if ((turtle_x < turtle_left_max + turtle_height) &&
 	    (current_mode == wrapmode)) {
 	turtle_x += (screen_width + 1);
@@ -150,11 +159,10 @@ check_x_low()
     }
 }
 
-draw_turtle_helper()
-{
+void draw_turtle_helper(void) {
     pen_info saved_pen;
     FLONUM real_heading;
-    int left_x, left_y, right_x, right_y, top_x, top_y, center_x, center_y;
+    int left_x, left_y, right_x, right_y, top_x, top_y;
     
     prepare_to_draw;
     prepare_to_draw_turtle;
@@ -186,22 +194,24 @@ draw_turtle_helper()
 }
 
 /************************************************************/
+void forward_helper(FLONUM);
+BOOLEAN safe_to_save(void);
+void save_line(void), save_move(void), save_vis(void), save_mode(void);
+void save_color(void), save_size(void), save_pattern(void);
+void save_string(char *, int);
 
-void right(FLONUM a)
-{
+void right(FLONUM a) {
     draw_turtle();
     turtle_heading += a;
     turtle_heading = pfmod(turtle_heading,360.0);
     draw_turtle();
 }
 
-NODE *numeric_arg(NODE *args)
-{
+NODE *numeric_arg(NODE *args) {
     NODE *arg = car(args), *val;
 
     val = cnv_node_to_numnode(arg);
     while (val == UNBOUND && NOT_THROWING) {
-	gcref(val);
 	setcar(args, err_logo(BAD_DATA, arg));
 	arg = car(args);
 	val = cnv_node_to_numnode(arg);
@@ -210,8 +220,7 @@ NODE *numeric_arg(NODE *args)
     return(val);
 }
 
-NODE *lright(NODE *arg)
-{
+NODE *lright(NODE *arg) {
     NODE *val;
     FLONUM a;
     
@@ -226,8 +235,7 @@ NODE *lright(NODE *arg)
     return(UNBOUND);
 }
 
-NODE *lleft(NODE *arg)
-{
+NODE *lleft(NODE *arg) {
     NODE *val;
     FLONUM a;
     
@@ -242,8 +250,12 @@ NODE *lleft(NODE *arg)
     return(UNBOUND);
 }
 
-void forward(FLONUM d)
-{
+int wrap_right(FLONUM, FLONUM, FLONUM, FLONUM, FLONUM);
+int wrap_left(FLONUM, FLONUM, FLONUM, FLONUM, FLONUM);
+int wrap_up(FLONUM, FLONUM, FLONUM, FLONUM, FLONUM);
+int wrap_down(FLONUM, FLONUM, FLONUM, FLONUM, FLONUM);
+
+void forward(FLONUM d) {
     prepare_to_draw;
     draw_turtle();
     forward_helper(d);
@@ -254,15 +266,16 @@ void forward(FLONUM d)
     out_of_bounds = FALSE;
 }
 
-forward_helper(FLONUM d)
-{
-    FLONUM real_heading, dx, dy, x1, y1, x2, y2, intercept;
+void forward_helper(FLONUM d) {
+    FLONUM real_heading, dx, dy, x1, y1, x2, y2;
     
     real_heading = -turtle_heading + 90.0;
     x1 = screen_x_coord;
     y1 = screen_y_coord;
     dx = (FLONUM)(cos((FLONUM)(real_heading*degrad))*d*x_scale);
     dy = (FLONUM)(sin((FLONUM)(real_heading*degrad))*d*y_scale);
+    if ((dx < 0 && dx > -0.000001) || (dx > 0 && dx < 0.000001)) dx = 0;
+    if ((dy < 0 && dy > -0.000001) || (dy > 0 && dy < 0.000001)) dy = 0;
     x2 = x1 + dx;
     y2 = y1 - dy;
     
@@ -273,7 +286,12 @@ forward_helper(FLONUM d)
     }
     
     if (check_throwing) return;
-    
+
+    if (internal_penmode == PENMODE_REVERSE && pen_vis == 0 && d > 0.0) {
+	line_to(g_round(x1), g_round(y1));	/* flip the corner */
+	save_line();
+    }
+
     if (current_mode == windowmode ||
 	(x2 >= screen_left && x2 <= screen_right &&
 	 y2 >= screen_top && y2 <= screen_bottom)) {
@@ -287,14 +305,18 @@ forward_helper(FLONUM d)
 	    if (!wrap_left(d, x1, y1, x2, y2))
 		if (!wrap_up(d, x1, y1, x2, y2))
 		    wrap_down(d, x1, y1, x2, y2);
+
+    if (internal_penmode == PENMODE_REVERSE && pen_vis == 0 && d < 0.0) {
+	line_to(g_round(screen_x_coord), g_round(screen_y_coord));
+	save_line();
+    }
 }
 
-int wrap_right(FLONUM d, FLONUM x1, FLONUM y1, FLONUM x2, FLONUM y2)
-{
+int wrap_right(FLONUM d, FLONUM x1, FLONUM y1, FLONUM x2, FLONUM y2) {
     FLONUM yi, newd;
     
     if (x2 > screen_right) {
-	yi = ((y2 - y1)/(x2 - x1)) * (screen_right - x1) + y1;
+	yi = ((y2 - y1)/(x2 - x1)) * (screen_right + 1 - x1) + y1;
 	if (yi >= screen_top && yi <= screen_bottom) {
 	    line_to(screen_right, g_round(yi));
 	    save_line();
@@ -313,12 +335,11 @@ int wrap_right(FLONUM d, FLONUM x1, FLONUM y1, FLONUM x2, FLONUM y2)
     return(0);
 }
 
-int wrap_left(FLONUM d, FLONUM x1, FLONUM y1, FLONUM x2, FLONUM y2)
-{
+int wrap_left(FLONUM d, FLONUM x1, FLONUM y1, FLONUM x2, FLONUM y2) {
     FLONUM yi, newd;
     
     if (x2 < screen_left) {
-	yi = ((y1 - y2)/(x2 - x1)) * (x1 - screen_left) + y1;
+	yi = ((y1 - y2)/(x2 - x1)) * (x1 + 1 - screen_left) + y1;
 	if (yi >= screen_top && yi <= screen_bottom) {
 	    line_to(screen_left, g_round(yi));
 	    save_line();
@@ -337,12 +358,11 @@ int wrap_left(FLONUM d, FLONUM x1, FLONUM y1, FLONUM x2, FLONUM y2)
     return(0);
 }
 
-int wrap_up(FLONUM d, FLONUM x1, FLONUM y1, FLONUM x2, FLONUM y2)
-{
+int wrap_up(FLONUM d, FLONUM x1, FLONUM y1, FLONUM x2, FLONUM y2) {
     FLONUM xi, newd;
     
     if (y2 < screen_top) {
-	xi = ((x2 - x1)/(y1 - y2)) * (y1 - screen_top) + x1;
+	xi = ((x2 - x1)/(y1 - y2)) * (y1 + 1 - screen_top) + x1;
 	if (xi >= screen_left && xi <= screen_right) {
 	    line_to(g_round(xi), screen_top);
 	    save_line();
@@ -361,12 +381,11 @@ int wrap_up(FLONUM d, FLONUM x1, FLONUM y1, FLONUM x2, FLONUM y2)
     return(0);
 }
 
-int wrap_down(FLONUM d, FLONUM x1, FLONUM y1, FLONUM x2, FLONUM y2)
-{
+int wrap_down(FLONUM d, FLONUM x1, FLONUM y1, FLONUM x2, FLONUM y2) {
     FLONUM xi, newd;
     
     if (y2 > screen_bottom) {
-	xi = ((x2 - x1)/(y2 - y1)) * (screen_bottom - y1) + x1;
+	xi = ((x2 - x1)/(y2 - y1)) * (screen_bottom + 1 - y1) + x1;
 	if (xi >= screen_left && xi <= screen_right) {
 	    line_to(g_round(xi), screen_bottom);
 	    save_line();
@@ -385,8 +404,7 @@ int wrap_down(FLONUM d, FLONUM x1, FLONUM y1, FLONUM x2, FLONUM y2)
     return(0);
 }
 
-NODE *lforward(NODE *arg)
-{
+NODE *lforward(NODE *arg) {
     NODE *val;
     FLONUM d;
     
@@ -401,8 +419,7 @@ NODE *lforward(NODE *arg)
     return(UNBOUND);
 }
 
-NODE *lback(NODE *arg)
-{
+NODE *lback(NODE *arg) {
     NODE *val;
     FLONUM d;
     
@@ -417,8 +434,7 @@ NODE *lback(NODE *arg)
     return(UNBOUND);
 }
 
-NODE *lshowturtle()
-{
+NODE *lshowturtle(NODE *args) {
     prepare_to_draw;
     if (!turtle_shown) {
 	turtle_shown = TRUE;
@@ -428,8 +444,7 @@ NODE *lshowturtle()
     return(UNBOUND);
 }
 
-NODE *lhideturtle()
-{
+NODE *lhideturtle(NODE *args) {
     prepare_to_draw;
     if (turtle_shown) {
 	draw_turtle();
@@ -439,13 +454,11 @@ NODE *lhideturtle()
     return(UNBOUND);
 }
 
-NODE *lshownp()
-{
+NODE *lshownp(NODE *args) {
     return(turtle_shown ? True : False);
 }
 
-NODE *lsetheading(NODE *arg)
-{
+NODE *lsetheading(NODE *arg) {
     NODE *val;
     
     val = numeric_arg(arg);
@@ -461,14 +474,12 @@ NODE *lsetheading(NODE *arg)
     return(UNBOUND);
 }
 
-NODE *lheading()
-{
+NODE *lheading(NODE *args) {
     return(make_floatnode(turtle_heading));
 }
 
-NODE *vec_arg_helper(NODE *args, BOOLEAN floatok, BOOLEAN three)
-{
-    NODE *arg = car(args), *val1, *val2, *val3;
+NODE *vec_arg_helper(NODE *args, BOOLEAN floatok, BOOLEAN three) {
+    NODE *arg = car(args), *val1, *val2, *val3 = NIL;
 
     while (NOT_THROWING) {
 	if (arg != NIL &&
@@ -481,15 +492,12 @@ NODE *vec_arg_helper(NODE *args, BOOLEAN floatok, BOOLEAN three)
 	    if (val1 != UNBOUND && val2 != UNBOUND &&
 		(floatok || (nodetype(val1) == INT && getint(val1) >= 0 &&
 			     nodetype(val2) == INT && getint(val2) >= 0 &&
-				 (!three || nodetype(val3) == INT && getint(val3) >= 0)))) {
+				 (!three || (nodetype(val3) == INT && getint(val3) >= 0))))) {
 		setcar(arg, val1);
 		setcar(cdr(arg), val2);
 		if (three) setcar (cddr(arg), val3);
 		return(arg);
 	    }
-	    gcref(val1);
-	    gcref(val2);
-		if (three) gcref(val3);
 	}
 	setcar(args, err_logo(BAD_DATA, arg));
 	arg = car(args);
@@ -509,8 +517,7 @@ NODE *rgb_arg(NODE *args) {
 	return vec_arg_helper(args,FALSE,TRUE);
 }
 
-FLONUM towards_helper(FLONUM x, FLONUM y, FLONUM from_x, FLONUM from_y)
-{
+FLONUM towards_helper(FLONUM x, FLONUM y, FLONUM from_x, FLONUM from_y) {
     FLONUM m, a, tx, ty;
 
     tx = from_x/x_scale;
@@ -527,10 +534,10 @@ FLONUM towards_helper(FLONUM x, FLONUM y, FLONUM from_x, FLONUM from_y)
 	a = -(a - 90.0);
 	return (a < 0 ? 360.0+a : a);
     }
+    return 0.0;
 }
 
-NODE *ltowards(NODE *args)
-{
+NODE *ltowards(NODE *args) {
     NODE *xnode, *ynode = UNBOUND, *arg;
     FLONUM x, y;
     
@@ -539,45 +546,40 @@ NODE *ltowards(NODE *args)
 	xnode = car(arg);
 	ynode = cadr(arg);
 	
-	x = ((nodetype(xnode) == FLOAT) ? getfloat(xnode) :
+	x = ((nodetype(xnode) == FLOATT) ? getfloat(xnode) :
 			  (FLONUM)getint(xnode));
-	y = ((nodetype(ynode) == FLOAT) ? getfloat(ynode) :
+	y = ((nodetype(ynode) == FLOATT) ? getfloat(ynode) :
 			  (FLONUM)getint(ynode));
 	return make_floatnode(towards_helper(x, y, turtle_x, turtle_y));
     }
     return(UNBOUND);
 }
 
-NODE *lpos()
-{
+NODE *lpos(NODE *args) {
     return(cons(make_floatnode(cut_error(turtle_x/x_scale)),
 	cons(make_floatnode(cut_error(turtle_y/y_scale)), NIL)));
 }
 
-NODE *lscrunch()
-{
+NODE *lscrunch(NODE *args) {
     return(cons(make_floatnode(x_scale), cons(make_floatnode(y_scale), NIL)));
 }
 
-NODE *lhome()
-{
-    void setpos_helper(NODE *, NODE *);
-
+NODE *lhome(NODE *args) {
     out_of_bounds = FALSE;
-    setpos_helper(make_intnode((FIXNUM)0), make_intnode((FIXNUM)0));
+    setpos_bynumber((FLONUM)0.0, (FLONUM)0.0);
     draw_turtle();
     turtle_heading = 0.0;
     draw_turtle();
     return(UNBOUND);
 }
 
-cs_helper(int centerp)
-{    
+void cs_helper(int centerp) {    
     prepare_to_draw;
     clear_screen;
     if (centerp) {
 	wanna_x = wanna_y = turtle_x = turtle_y = turtle_heading = 0.0;
 	out_of_bounds = FALSE;
+	move_to(screen_x_coord, screen_y_coord);
     }
     draw_turtle();
     save_pen(&orig_pen);
@@ -587,35 +589,21 @@ cs_helper(int centerp)
     done_drawing;
 }
 
-NODE *lclearscreen()
-{
+NODE *lclearscreen(NODE *args) {
     cs_helper(TRUE);
     return(UNBOUND);
 }
 
-NODE *lclean()
-{
+NODE *lclean(NODE *args) {
     cs_helper(FALSE);
     return(UNBOUND);
 }
 
-void setpos_helper(NODE *xnode, NODE *ynode)
-{
-    FLONUM target_x, target_y, scaled_x, scaled_y, tx, ty, save_heading;
+void setpos_commonpart(FLONUM target_x, FLONUM target_y) {
+    FLONUM scaled_x, scaled_y, tx, ty, save_heading;
     BOOLEAN wrapping = FALSE;
     
     if (NOT_THROWING) {
-	prepare_to_draw;
-	draw_turtle();
-	move_to(g_round(screen_x_coord), g_round(screen_y_coord));
-	target_x = ((xnode == NIL) ?
-		turtle_x :
-		((nodetype(xnode) == FLOAT) ? getfloat(xnode) :
-		 (FLONUM)getint(xnode)));
-	target_y = ((ynode == NIL) ?
-		turtle_y :
-		((nodetype(ynode) == FLOAT) ? getfloat(ynode) :
-		 (FLONUM)getint(ynode)));
 	scaled_x = target_x * x_scale;
 	scaled_y = target_y * y_scale;
 	wrapping = scaled_x > turtle_right_max || scaled_x < turtle_left_max ||
@@ -636,10 +624,11 @@ void setpos_helper(NODE *xnode, NODE *ynode)
 	    out_of_bounds = wrapping;
 	}
 	else {
+	    out_of_bounds = FALSE;
 	    wanna_x = turtle_x = scaled_x;
 	    wanna_y = turtle_y = scaled_y;
-	    out_of_bounds = FALSE;
-	    line_to(g_round(screen_x_coord), g_round(screen_y_coord));
+	    line_to(g_round(screen_x_coord),
+		    g_round(screen_y_coord));
 	    save_line();
 	}
 	done_drawing;
@@ -647,8 +636,35 @@ void setpos_helper(NODE *xnode, NODE *ynode)
     }
 }
 
-NODE *lsetpos(NODE *args)
-{
+void setpos_bynumber(FLONUM target_x, FLONUM target_y) {
+    if (NOT_THROWING) {
+	prepare_to_draw;
+	draw_turtle();
+	move_to(g_round(screen_x_coord), g_round(screen_y_coord));
+	setpos_commonpart(target_x, target_y);
+    }
+}
+
+void setpos_helper(NODE *xnode, NODE *ynode) {
+    FLONUM target_x, target_y;
+    
+    if (NOT_THROWING) {
+	prepare_to_draw;
+	draw_turtle();
+	move_to(g_round(screen_x_coord), g_round(screen_y_coord));
+	target_x = ((xnode == NIL) ?
+		turtle_x :
+		((nodetype(xnode) == FLOATT) ? getfloat(xnode) :
+		 (FLONUM)getint(xnode)));
+	target_y = ((ynode == NIL) ?
+		turtle_y :
+		((nodetype(ynode) == FLOATT) ? getfloat(ynode) :
+		 (FLONUM)getint(ynode)));
+	setpos_commonpart(target_x, target_y);
+    }
+}
+
+NODE *lsetpos(NODE *args) {
     NODE *arg = vector_arg(args);
 
     if (NOT_THROWING) {
@@ -657,8 +673,7 @@ NODE *lsetpos(NODE *args)
     return(UNBOUND);
 }
 
-NODE *lsetxy(NODE *args)
-{
+NODE *lsetxy(NODE *args) {
     NODE *xnode, *ynode;
     
     xnode = numeric_arg(args);
@@ -669,8 +684,7 @@ NODE *lsetxy(NODE *args)
     return(UNBOUND);
 }
 
-NODE *lsetx(NODE *args)
-{
+NODE *lsetx(NODE *args) {
     NODE *xnode;
     
     xnode = numeric_arg(args);
@@ -680,8 +694,7 @@ NODE *lsetx(NODE *args)
     return(UNBOUND);
 }
 
-NODE *lsety(NODE *args)
-{
+NODE *lsety(NODE *args) {
     NODE *ynode;
     
     ynode = numeric_arg(args);
@@ -691,44 +704,53 @@ NODE *lsety(NODE *args)
     return(UNBOUND);
 }
 
-NODE *lwrap()
-{
-    if (turtle_shown) draw_turtle();
+NODE *lwrap(NODE *args) {
+    draw_turtle();
     current_mode = wrapmode;
-    if (turtle_shown) draw_turtle();
+    while (turtle_x > turtle_right_max) {
+	turtle_x -= screen_width;
+    }
+    while (turtle_x < turtle_left_max) {
+	turtle_x += screen_width;
+    }
+    while (turtle_y > turtle_top_max) {
+	turtle_y -= screen_height;
+    }
+    while (turtle_y < turtle_bottom_max) {
+	turtle_y += screen_height;
+    }
+    move_to(screen_x_coord, screen_y_coord);
+    draw_turtle();
     return(UNBOUND);
 }
 
-NODE *lfence()
-{
-    if (turtle_shown) draw_turtle();
+NODE *lfence(NODE *args) {
+    (void)lwrap(args);	    /* get turtle inside the fence */
+    draw_turtle();
     current_mode = fencemode;
-    if (turtle_shown) draw_turtle();
+    draw_turtle();
     return(UNBOUND);
 }
 
-NODE *lwindow()
-{
-    if (turtle_shown) draw_turtle();
+NODE *lwindow(NODE *args) {
+    draw_turtle();
     current_mode = windowmode;
-    if (turtle_shown) draw_turtle();
+    draw_turtle();
     return(UNBOUND);
 }
 
-NODE *lfill()
-{
-    if (turtle_shown) draw_turtle();
+NODE *lfill(NODE *args) {
+    draw_turtle();
     logofill();
-    if (turtle_shown) draw_turtle();
+    draw_turtle();
     if (safe_to_save()) {
 	record[record_index] = FILLERUP;
-	record_index += 2;
+	record_index += One;
     }
     return(UNBOUND);
 }
 
-NODE *llabel(NODE *arg)
-{
+NODE *llabel(NODE *arg) {
     char textbuf[300];
     short theLength;
 
@@ -738,127 +760,114 @@ NODE *llabel(NODE *arg)
     *print_stringptr = '\0';
 	
     if (NOT_THROWING) {
+	prepare_to_draw;
 	draw_turtle();
-#ifdef x_window
-	label(textbuf, strlen(textbuf));
-#else
-#ifdef mac
 	theLength = strlen(textbuf);
+#ifdef mac
 	c_to_pascal_string(textbuf, theLength);
 #endif
 	label(textbuf);
-	save_string(textbuf);
+	save_string(textbuf,theLength);
 	record_next_move = TRUE;
-#endif
 	draw_turtle();
+	done_drawing;
     }
     return(UNBOUND);
 }
 
-NODE *ltextscreen()
-{
+NODE *ltextscreen(NODE *args) {
     text_screen;
     return(UNBOUND);
 }
 
-NODE *lsplitscreen()
-{
+NODE *lsplitscreen(NODE *args) {
     split_screen;
     return(UNBOUND);
 }
 
-NODE *lfullscreen()
-{
+NODE *lfullscreen(NODE *args) {
     full_screen;
     return(UNBOUND);
 }
 
-NODE *lpendownp()
-{
+NODE *lpendownp(NODE *args) {
     return(pen_vis == 0 ? True : False);
 }
 
-NODE *lpenmode()
-{
+NODE *lpenmode(NODE *args) {
     return(get_node_pen_mode);
 }
 
-NODE *lpencolor()
-{
+NODE *lpencolor(NODE *args) {
     return(make_intnode((FIXNUM)pen_color));
 }
 
-NODE *lbackground()
-{
+NODE *lbackground(NODE *args) {
     return(make_intnode((FIXNUM)back_ground));
 }
 
-NODE *lpensize()
-{
+NODE *lpensize(NODE *args) {
     return(cons(make_intnode((FIXNUM)pen_width),
 	cons(make_intnode((FIXNUM)pen_height), NIL)));
 }
 
-NODE *lpenpattern()
-{
+NODE *lpenpattern(NODE *args) {
     return(get_node_pen_pattern);
 }
 
-NODE *lpendown()
-{
+NODE *lpendown(NODE *args) {
     pen_vis = 0;
     save_vis();
     return(UNBOUND);
 }
 
-NODE *lpenup()
-{
+NODE *lpenup(NODE *args) {
     if (pen_vis == 0)
 	pen_vis--;
     save_vis();
     return(UNBOUND);
 }
 
-NODE *lpenpaint()
-{
+NODE *lpenpaint(NODE *args) {
+    internal_penmode = PENMODE_PAINT;
     pen_down;
     save_mode();
-    return(lpendown());
+    return(lpendown(NIL));
 }
 
-NODE *lpenerase()
-{
+NODE *lpenerase(NODE *args) {
+    internal_penmode = PENMODE_ERASE;
     pen_erase;
     save_mode();
-    return(lpendown());
+    return(lpendown(NIL));
 }
 
-NODE *lpenreverse()
-{
+NODE *lpenreverse(NODE *args) {
+    internal_penmode = PENMODE_REVERSE;
     pen_reverse;
     save_mode();
-    return(lpendown());
+    return(lpendown(NIL));
 }
 
-NODE *lsetpencolor(NODE *arg)
-{
+NODE *lsetpencolor(NODE *arg) {
     NODE *val = pos_int_arg(arg);
 
     if (NOT_THROWING) {
 	prepare_to_draw;
 	set_pen_color(getint(val));
 	save_color();
+	done_drawing;
     }
     return(UNBOUND);
 }
 
-NODE *lsetbackground(NODE *arg)
-{
+NODE *lsetbackground(NODE *arg) {
     NODE *val = pos_int_arg(arg);
 
     if (NOT_THROWING) {
 	prepare_to_draw;
 	set_back_ground(getint(val));
+	done_drawing;
     }
     return(UNBOUND);
 }
@@ -868,10 +877,12 @@ NODE *lsetpalette(NODE *args) {
 	NODE *arg = rgb_arg(cdr(args));
 
 	if (NOT_THROWING && ((int)getint(slot) > 7)) {
+		prepare_to_draw;
 		set_palette((int)getint(slot),
-					(unsigned int)getint(car(arg)),
-					(unsigned int)getint(cadr(arg)),
-					(unsigned int)getint(car(cddr(arg))));
+			    (unsigned int)getint(car(arg)),
+			    (unsigned int)getint(cadr(arg)),
+			    (unsigned int)getint(car(cddr(arg))));
+		done_drawing;
 	}
 	return(UNBOUND);
 }
@@ -889,8 +900,7 @@ NODE *lpalette(NODE *args) {
 	return UNBOUND;
 }
 
-NODE *lsetpensize(NODE *args)
-{
+NODE *lsetpensize(NODE *args) {
     NODE *arg = pos_int_vector_arg(args);
 
     if (NOT_THROWING) {
@@ -898,31 +908,29 @@ NODE *lsetpensize(NODE *args)
 	set_pen_width((int)getint(car(arg)));
 	set_pen_height((int)getint(cadr(arg)));
 	save_size();
+	done_drawing;
     }
     return(UNBOUND);
 }
 
-NODE *lsetpenpattern(NODE *args)
-{    
+NODE *lsetpenpattern(NODE *args) {    
     NODE *arg;
 
     arg = car(args);
-    ref(arg);
-    while ((arg == NIL || !is_list(arg)) && NOT_THROWING)
-	arg = reref(arg, err_logo(BAD_DATA, arg));
+    while ((!is_list(arg)) && NOT_THROWING)
+	arg = err_logo(BAD_DATA, arg);
 	
     if (NOT_THROWING) {
 	prepare_to_draw;
 	set_list_pen_pattern(arg);
 	save_pattern();
+	done_drawing;
     }
 
-    deref(arg);
     return(UNBOUND);
 }
 
-NODE *lsetscrunch(NODE *args)
-{
+NODE *lsetscrunch(NODE *args) {
     NODE *xnode, *ynode;
 
     xnode = numeric_arg(args);
@@ -930,24 +938,24 @@ NODE *lsetscrunch(NODE *args)
 
     if (NOT_THROWING) {
 	prepare_to_draw;
-	if (turtle_shown) {
-	    draw_turtle();
-    	}
-	x_scale = (nodetype(xnode) == FLOAT) ? getfloat(xnode) :
+	draw_turtle();
+	x_scale = (nodetype(xnode) == FLOATT) ? getfloat(xnode) :
 			       (FLONUM)getint(xnode);
-	y_scale = (nodetype(ynode) == FLOAT) ? getfloat(ynode) :
+	y_scale = (nodetype(ynode) == FLOATT) ? getfloat(ynode) :
 			       (FLONUM)getint(ynode);
-	if (turtle_shown) {
-	    draw_turtle();
-    	}
+	draw_turtle();
 	done_drawing;
 #ifdef __ZTC__
 	{
-	    FILE *fp = fopen("scrunch.dat","w");
+	    FILE *fp = fopen("scrunch.dat","r");
 	    if (fp != NULL) {
-		fwrite(&x_scale, sizeof(FLONUM), 1, fp);
-		fwrite(&y_scale, sizeof(FLONUM), 1, fp);
 		fclose(fp);
+		fp = fopen("scrunch.dat","w");
+		if (fp != NULL) {
+		    fwrite(&x_scale, sizeof(FLONUM), 1, fp);
+		    fwrite(&y_scale, sizeof(FLONUM), 1, fp);
+		    fclose(fp);
+		}
 	    }
 	}
 #endif
@@ -955,21 +963,22 @@ NODE *lsetscrunch(NODE *args)
     return(UNBOUND);
 }
 
-NODE *lmousepos()
-{
+NODE *lmousepos(NODE *args) {
+#ifdef WIN32 /* sowings */
+    return NIL;
+#else
     return(cons(make_intnode(mouse_x), cons(make_intnode(mouse_y), NIL)));
+#endif
 }
 
-NODE *lbuttonp()
-{
+NODE *lbuttonp(NODE *args) {
     if (button)
 	return(True);
     return(False);
 }
 
-NODE *ltone(args)
-NODE *args;
-{
+NODE *ltone(NODE *args) {
+#ifndef WIN32
     NODE *p, *d;
     FIXNUM pitch, duration;
     
@@ -977,16 +986,18 @@ NODE *args;
     d = numeric_arg(cdr(args));
     
     if (NOT_THROWING) {
-	pitch = (nodetype(p) == FLOAT) ? (FIXNUM)getfloat(p) : getint(p);
-	duration = (nodetype(d) == FLOAT) ? (FIXNUM)getfloat(d) : getint(d);
+	pitch = (nodetype(p) == FLOATT) ? (FIXNUM)getfloat(p) : getint(p);
+	duration = (nodetype(d) == FLOATT) ? (FIXNUM)getfloat(d) : getint(d);
 	if (pitch > 0) tone(pitch, duration);
     }
+#else
+    MessageBeep(0xffffffff);
+#endif
     
     return(UNBOUND);
 }
 
-NODE *larc(NODE *arg)
-{
+NODE *larc(NODE *arg) {
     NODE *val1;
     NODE *val2;
 
@@ -995,8 +1006,6 @@ NODE *larc(NODE *arg)
     FLONUM ang;
     FLONUM tx;
     FLONUM ty;
-    FLONUM oldx;
-    FLONUM oldy;
     FLONUM x;
     FLONUM y;
     FLONUM count;
@@ -1060,7 +1069,7 @@ NODE *larc(NODE *arg)
           pen_state = pen_vis;
           pen_vis = -1;
 	  save_vis();
-          setpos_helper(make_floatnode(tx+x), make_floatnode(ty+y));
+          setpos_bynumber(tx+x, ty+y);
           pen_vis = pen_state;
 	  save_vis();
           }
@@ -1069,7 +1078,7 @@ NODE *larc(NODE *arg)
 
        else
           {
-          setpos_helper(make_floatnode(tx+x), make_floatnode(ty+y));
+          setpos_bynumber(tx+x, ty+y);
           }
 
        ang = ang + delta;
@@ -1080,7 +1089,7 @@ NODE *larc(NODE *arg)
     x = sin((turtle_heading+angle)*3.141592654/180.0)*radius;
     y = cos((turtle_heading+angle)*3.141592654/180.0)*radius;
 
-    setpos_helper(make_floatnode(tx+x), make_floatnode(ty+y));
+    setpos_bynumber(tx+x, ty+y);
 
     /* restore state */
 
@@ -1090,11 +1099,17 @@ NODE *larc(NODE *arg)
     turtle_y = ty;
 
     draw_turtle();
-    done_drawing;
     wanna_x = turtle_x;
     wanna_y = turtle_y;
     out_of_bounds = FALSE;
+    pen_state = pen_vis;
+    pen_vis = -1;
+    save_vis();
+    forward_helper((FLONUM)0.0);    /* Lets fill work -- dunno why */
+    pen_vis = pen_state;
+    save_vis();
 
+    done_drawing;
     }
     return(UNBOUND);
 }
@@ -1107,189 +1122,190 @@ NODE *larc(NODE *arg)
    machines where no graphics redrawing is necessary, set the size
    of the recording buffer to 1 in logo.h. */
 
-BOOLEAN safe_to_save()
-{
+BOOLEAN safe_to_save(void) {
     return(refresh_p && record_index < (GR_SIZE - 300));
 }
 
-save_lm_helper ()
-{
-    *(int *)(record + record_index + 2) = pen_x;
-    *(int *)(record + record_index + 4) = pen_y;
-    record_index += 6;
+void save_lm_helper (void) {
+    *(int *)(record + record_index + One) = pen_x;
+    *(int *)(record + record_index + Two) = pen_y;
+    record_index += Three;
 }
 
-save_line()
-{
+void save_line(void) {
     if (safe_to_save()) {
 	record[record_index] = LINEXY;
 	save_lm_helper();
     }
 }
 
-save_move()
-{
+void save_move(void) {
     if (safe_to_save()) {
 	record[record_index] = MOVEXY;
 	save_lm_helper();
     }
 }
 
-save_vis()
-{
+void save_vis(void) {
     if (safe_to_save()) {
 	record[record_index] = SETPENVIS;
 	record[record_index + 1] = (char)pen_vis;
-	record_index += 2;
+	record_index += One;
     }
 }
 
-save_mode()
-{
+void save_mode(void) {
     if (safe_to_save()) {
 	record[record_index] = SETPENMODE;
 #ifdef x_window
-	*(GC *)(record + record_index + 2) = pen_mode;
+	*(GC *)(record + record_index + One) = pen_mode;
 #else
-	*(int *)(record + record_index + 2) = pen_mode;
+	*(int *)(record + record_index + One) = pen_mode;
 #endif
-	record_index += 4;
+	*(int *)(record + record_index + Two) = internal_penmode;
+	record_index += Three;
 	save_color();
     }
 }
 
-save_color()
-{
+void save_color(void) {
     if (safe_to_save()) {
 	record[record_index] = SETPENCOLOR;
-	*(long *)(record + record_index + 2) = pen_color;
-	record_index += 6;
+	*(int *)(record + record_index + One) = pen_color;
+	record_index += Two;
     }
 }
 
-save_size()
-{
+void save_size(void) {
     if (safe_to_save()) {
 	record[record_index] = SETPENSIZE;
-	*(int *)(record + record_index + 2) = pen_width;
-	*(int *)(record + record_index + 4) = pen_height;
-	record_index += 6;
+	*(int *)(record + record_index + One) = pen_width;
+	*(int *)(record + record_index + Two) = pen_height;
+	record_index += Three;
     }
 }
 
-save_pattern()
-{
-    int count;
-    
+void save_pattern(void) {
     if (safe_to_save()) {
 	record[record_index] = SETPENPATTERN;
-	get_pen_pattern(&record[record_index + 2]);
-	record_index += 10;
+	get_pen_pattern(&record[record_index + One]);
+	record_index += One+8;
     }
 }
 
-save_string(s)
-char s[];
-{
+void save_string(char *s, int len) {
     int count;
 
     if (safe_to_save()) {
 	record[record_index] = LABEL;
-	record[record_index + 2] = s[0];
-	for (count = 0; count < s[0]; count++)
-	    record[record_index + 3 + count] = s[1 + count];
-	record_index += 3 + s[0] + even_p(s[0]);
+	record[record_index + One] = (unsigned char)len;
+	for (count = 0; count <= len; count++)
+	    record[record_index + One+1 + count] = s[count];
+	record[record_index + One+2 + len] = '\0';
+	record_index += (One+2 + len + (One-1)) & ~(One-1);
     }
 }
 
-NODE *lrefresh()
-{
+NODE *lrefresh(NODE *args) {
     refresh_p = TRUE;
     return(UNBOUND);
 }
 
-NODE *lnorefresh()
-{
+NODE *lnorefresh(NODE *args) {
     refresh_p = FALSE;
     return(UNBOUND);
 }
 
-void redraw_graphics()
-{
+void redraw_graphics(void) {
     int r_index = 0;
-    /* pen_info saved_pen; */
+    pen_info saved_pen;
     BOOLEAN saved_shown;
+#if defined(__ZTC__) && !defined(WIN32)
+    BOOLEAN save_splitscreen = in_splitscreen;
+#endif
     
     if (!refresh_p) {
-	draw_turtle();
+    /*	clear_screen;
+	draw_turtle();	*/
 	return;
     }
 
     saved_shown = turtle_shown;
     turtle_shown = FALSE;
-    /* save_pen(&saved_pen); */
+    save_pen(&saved_pen);
     restore_pen(&orig_pen);
 
-#ifndef unix
-    erase_screen();
+#if defined(__ZTC__) && !defined(WIN32)
+    full_screen;
 #endif
+
+    erase_screen();
+
 #ifdef __TURBOC__
     moveto(p_info_x(orig_pen),p_info_y(orig_pen));
 #endif
+
     while (r_index < record_index)
 	switch (record[r_index]) {
 	    case (LINEXY) :
-		line_to(*(int *)(record + r_index + 2), *(int *)(record + r_index + 4));
-		r_index += 6;
+		line_to(*(int *)(record + r_index + One),
+			*(int *)(record + r_index + Two));
+		r_index += Three;
 		break;
 	    case (MOVEXY) :
-		move_to(*(int *)(record + r_index + 2), *(int *)(record + r_index + 4));
-		r_index += 6;
+		move_to(*(int *)(record + r_index + One),
+			*(int *)(record + r_index + Two));
+		r_index += Three;
 		break;
 	    case (LABEL) :
-		draw_string(record + r_index + 2);
-		r_index += 3 + record[r_index + 2] + even_p(record[r_index + 2]);
+		draw_string(record + r_index + One+1);
+		r_index += (One+2 + record[r_index + One] + (One-1)) & ~(One-1);
 		break;
 	    case (SETPENVIS) :
 		set_pen_vis(record[r_index + 1]);
-		r_index += 2;
+		r_index += One;
 		break;
 	    case (SETPENMODE) :
 #ifdef x_window
-		set_pen_mode(*(GC *)(record + r_index + 2));
+		set_pen_mode(*(GC *)(record + r_index + One));
 #else
-		set_pen_mode(*(int *)(record + r_index + 2));
+		set_pen_mode(*(int *)(record + r_index + One));
 #endif
-		r_index += 4;
+		internal_penmode = *(int *)(record + record_index + Two);
+		r_index += Three;
 		break;
 	    case (SETPENCOLOR) :
-		set_pen_color(*(long *)(record + r_index + 2));
-		r_index += 6;
+		set_pen_color(*(int *)(record + r_index + One));
+		r_index += Two;
 		break;
 	    case (SETPENSIZE) :
-		set_pen_width(*(int *)(record + r_index + 2));
-		set_pen_height(*(int *)(record + r_index + 4));
-		r_index += 6;
+		set_pen_width(*(int *)(record + r_index + One));
+		set_pen_height(*(int *)(record + r_index + Two));
+		r_index += Three;
 		break;
 	    case (SETPENPATTERN) :
-		set_pen_pattern(&record[r_index + 2]);
-		r_index += 10;
+		set_pen_pattern(&record[r_index + One]);
+		r_index += One+8;
 		break;
 	    case (FILLERUP) :
 		logofill();
-		r_index += 2;
+		r_index += One;
 		break;
 	}
 
-    /* restore_pen(&saved_pen); */
+    restore_pen(&saved_pen);
     turtle_shown = saved_shown;
+
+#if defined(__ZTC__) && !defined(WIN32)
+    if (save_splitscreen) {split_screen;}
+#endif
+
     draw_turtle();
 }
 
 /* This is called when the graphics coordinate system has been shifted.
    It adds a constant amount to each x and y coordinate in the record. */
-void resize_record(int dh, int dv)
-{
+void resize_record(int dh, int dv) {
     int r_index = 0;
     
     p_info_x(orig_pen) += dh;
@@ -1298,35 +1314,27 @@ void resize_record(int dh, int dv)
     while (r_index < record_index)
 	switch (record[r_index]) {
 	    case (LINEXY) :
-		*(int *)(record + r_index + 2) += dh;
-		*(int *)(record + r_index + 4) += dv;
-		r_index += 6;
-		break;
 	    case (MOVEXY) :
-		*(int *)(record + r_index + 2) += dh;
-		*(int *)(record + r_index + 4) += dv;
-		r_index += 6;
+		*(int *)(record + r_index + One) += dh;
+		*(int *)(record + r_index + Two) += dv;
+		r_index += Three;
 		break;
 	    case (LABEL) :
-		r_index += 3 + record[r_index + 2] + even_p(record[r_index + 2]);
+		r_index += (One+2 + record[r_index + One] + (One-1)) & ~(One-1);
 		break;
 	    case (SETPENVIS) :
-		r_index += 2;
-		break;
-	    case (SETPENMODE) :
-		r_index += 4;
+	    case (FILLERUP) :
+		r_index += One;
 		break;
 	    case (SETPENCOLOR) :
-		r_index += 6;
+		r_index += Two;
 		break;
 	    case (SETPENSIZE) :
-		r_index += 6;
+	    case (SETPENMODE) :
+		r_index += Three;
 		break;
 	    case (SETPENPATTERN) :
-		r_index += 10;
-		break;
-	    case (FILLERUP) :
-		r_index += 2;
+		r_index += One+8;
 		break;
 	}
 }
